#!/usr/bin/env python

from __future__ import nested_scopes

from exceptions import OSError
from cPickle import dumps, loads, Pickler, Unpickler
from getopt import getopt, GetoptError
from xml.parsers.expat import ExpatError
from mutex import mutex
from os import close,fork,pipe,read,write
from select import select
from signal import signal, SIGINT
from string import join
from sys import argv, exc_info, exit
from syslog import syslog, LOG_INFO, LOG_ERR
from time import sleep, strftime
from xml.dom.minidom import Document, parseString
from re import compile
from traceback import extract_tb

from sss.node import Node
from sss.server import SSSSelectServer
from sss.ssslib import comm_lib, ProtocolError, CommunicationError, ConnectError, InternalError, LookupError

class Event(Node):
    __attrfields__=['component','msg','data','time']

class Subscription(Node):
    __attrfields__=['component','msg','data','notify','matches','delivered','failures','last']
    regex_re = compile("r/(.*)/")

    def __str__(self):
        return "subscription: c=%s m=%s d=%s n=%s"%(self.component,self.msg,self.data,self.notify)

    def __setdefaults__(self):
        self.matches='0'
        self.delivered='0'
        self.failures='0'
        self.last='never'

    def __init__(self,xml):
        Node.__init__(self,xml)
        self.regex={}
        for field in [x for x in self.__attrfields__ if xml.hasAttribute(x)]:
            ere=self.regex_re.match(getattr(self,field))
            if ere: self.regex[field]=compile(ere.group(1))

    def EventMatch(self,event):
        for field in ['component','msg','data']:
            if self.regex.has_key(field):
                if not self.regex[field].match(getattr(event,field)):
                    return 0
            else:
                if (getattr(self,field) != getattr(event,field)) and (getattr(self,field) != '**'):
                    return 0
        return 1

    def __eq__(self,other):
        return ((self.notify == other.notify) and (self.component == other.component) and
                (self.data == other.data) and (self.msg == other.msg))

    def SetID(self,id):
        self.id=id

class event_cache:
    def __init__(self,min,max):
        self.cache=[]
        self.min=min
        self.max=max

    def Add(self,entry):
        self.cache.append(entry)
        if len(self.cache) > self.max:
            self.cache=self.cache[-self.min:]

    def __iter__(self):
        return iter(self.cache)

class emng(SSSSelectServer):
    __implementation__='emng'
    __component__='event-manager'
    
    def __parseinput__(self):
        self.subscriptions=[]
        self.statefile=self.kwargs["statefile"]
        try:
            (self.subid,self.subscriptions) = Unpickler(open(self.statefile)).load()
        except (IOError,OSError):
            pass
        self.cache=event_cache(600,2000)
        self.dispatch={'add-subscription':self.XMLSubscribe,
                       'del-subscription':self.XMLUnsubscribe,
                       'get-subscription':self.XMLGetSubscriptions,
                       'add-event':self.XMLEvent,
                       'get-event':self.XMLGetEvent}
        self.subid=0

    def GetSubID(self):
        self.subid=self.subid+1
        self.DumpData()
        return self.subid-1

    def ChildSigHand(self,signum,frame):
        exit(0)

    def HandleStatMsg(self,pipe):
        try:
            msg=self.PipeReadMsg(pipe)
        except OSError:
            syslog(LOG_ERR, "got broken pipe from child")
            syslog(LOG_ERR, "Statistics are offline")
            self.shut=1
        if msg.has_key('ssubids'):
            for subscription in self.subscriptions:
                if subscription.id in msg['ssubids']:
                    subscription.delivered=str(int(subscription.delivered)+msg['ssubids'].count(subscription.id))
                    subscription.last=strftime("%X %x")
        if msg.has_key('fsubids'):
            for subscription in self.subscriptions:
                if subscription.id in msg['fsubids']:
                    subscription.failures=str(int(subscription.failures)+msg['fsubids'].count(subscription.id))
        if msg.has_key('rsubids'):
            for subscription in self.subscriptions:
                if subscription.id in msg['rsubids']:
                    syslog(LOG_ERR, "Subscription for %s removed because of component death"%(subscription.notify))
                    self.subscriptions.remove(subscription)
            self.DumpData()
        return 0
        
    def __shutdown__(self):
        self.DumpData()

    def DeliverEvents(self,pipe,inl=None):
        inl=[pipe]
        while inl:
            (dest,event,subid)=self.PipeReadMsg(pipe)
            if self.todo.has_key(dest):
                self.todo[dest].append((event,subid))
            else:
                self.todo[dest]=[(event,subid)]
            # catch any other events coming down the pipe now
            (inl,foo,foo)=select([pipe],[],[],0)
        # Now we deliver contents of self.todo
        for component in self.todo.keys():
            stats={}
            try:
                h=self.comm.ClientInit(component)
                events=map(lambda x:x[0],self.todo[component])
                self.comm.SendMessage(h,"<events>%s</events>"%join(events,""))
                resp=self.comm.RecvMessage(h)
                self.comm.ClientClose(h)
                rdom=parseString(resp)
                if rdom.documentElement.tagName != 'event-ok':
                    syslog(LOG_ERR, "Failed to send event to %s"%(component))
                else:
                    stats['ssubids']=[x[1] for x in self.todo[component]]
                    del self.todo[component]
                            
            except ExpatError:
                syslog(LOG_ERR, "Bad response from %s"%(component))
                stats['fsubids']=[x[1] for x in self.todo[component]]
            except LookupError:
                syslog(LOG_ERR, "Component %s down. flushing queue"%(component))
                stats['rsubids']=[x[1] for x in self.todo[component]]
                del self.todo[component]
            except ConnectError:
                syslog(LOG_ERR, "Component %s down(connection refused). flushing queue"%(component))
                stats['rsubids']=[x[1] for x in self.todo[component]]
                del self.todo[component]
            except CommunicationError,m:
                syslog(LOG_ERR, "Component %s transmit failure: %s."%(component,m))
            except:
                (t,v,tb)=exc_info()
                errstr = ""
                for line in extract_tb(tb):
                    errstr += '  File "%s", line %i, in %s\n    %s\n'%line
                    errstr += "%s: %s\n"%(t,v)
                syslog(LOG_ERR, errstr)
            try:
                if stats: self.PipeWriteMsg(self.sw,stats)
            except OSError:
                syslog(LOG_ERR,"Pipe to parent closed, exiting")
                exit(1)
        return 0
        
    def __setupChildren__(self):
        # pw/pr are for events that need propagation
        (self.pr,self.pw)=pipe()
        # sw/sw are for statistics
        (self.sr,self.sw)=pipe()
        pid=fork()
        if pid == 0:
            self.todo={}
            signal(SIGINT, self.ChildSigHand)
            close(self.sr)
            self.readsockets=[self.pr]
            self.socketdispatch={self.pr:self.DeliverEvents}
            self.ServeForever()
        else:
            self.readsockets.append(self.sr)
            self.socketdispatch[self.sr]=self.HandleStatMsg
            close(self.pr)
            return [pid]

    def GetSources(self):
        sources=[]
        for sub in self.subscriptions:
            if ((sub.component not in sources) and (sub.component != "*")):
                sources.append(sub.component)
        return sources

    def GetSinks(self):
        sinks=[]
        for sub in self.subscriptions:
            if sub.notify not in sinks:
                sinks.append(sub.notify)
        return sinks

    def GetSubscriptions(self,dom):
        return filter(lambda x:x.Match(dom),self.subscriptions)

    def DelSubscriptions(self,dom):
        matching=self.GetSubscriptions(dom)
        map(lambda x:self.subscriptions.remove(x),matching)
        self.DumpData()
        return matching

    def AddEvent(self,event):
        self.cache.Add(event)
        for subscription in self.subscriptions:
            if subscription.EventMatch(event):
                subscription.matches=str(int(subscription.matches)+1)
                try:
                    self.PipeWriteMsg(self.pw,(subscription.notify,event.toxmldom().toxml(),subscription.id))
                except OSError, msg:
                    syslog(LOG_ERR, "Pipe write error %s"%str(msg))

    def XMLEvent(self,xml,(peer,port)):
        response=Document().createElement("events")
        for evdata in xml.getElementsByTagName("event"):
            self.AddEvent(Event(evdata))
            response.appendChild(evdata)
        return response

    def XMLGetEvent(self,xml,(peer,port)):
        response=Document().createElement("events")
        for spec in xml.childNodes:
            evspec=Event(spec)
            matching=map(lambda x:x.toxmldom(evspec),filter(lambda x:x.Match(evspec),self.cache))
            map(response.appendChild,matching)
        return response

    def XMLSubscribe(self,xml,(peer,port)):
        response=Document().createElement("subscriptions")
        for subdata in xml.childNodes:
            curr_sub=Subscription(subdata)
            # We need the subid to match statistics with subscriptions
            curr_sub.SetID(self.GetSubID())
            syslog(LOG_INFO,"Got sub from %s:%s"%(peer,curr_sub))
            if curr_sub not in self.subscriptions:
                self.subscriptions.append(curr_sub)
            response.appendChild(subdata)
        self.DumpData()
        return response

    def XMLUnsubscribe(self,xml,(peer,port)):
        response=Document().createElement("subscriptions")
        for usub in xml.childNodes:
            data=self.DelSubscriptions(Node(usub))
            map(lambda x:syslog(LOG_INFO,"Got unsubscribe from %s:%s"%(peer,x)),data)
            map(lambda x:response.appendChild(x.toxmldom()),data)
        self.DumpData()
        return response
        
    def XMLGetSubscriptions(self,xml,(peer,port)):
        response=Document().createElement("subscriptions")
        #response.setAttribute('total',str(len(self.subscriptions)))
        for usub in xml.childNodes:
            unsub=Node(usub)
            map(response.appendChild,map(lambda x:x.toxmldom(unsub),self.GetSubscriptions(unsub)))
        #response.setAttribute('matched',str(len(response.childNodes)))
        return response

    def DumpData(self):
        try:
            data = open(self.statefile,'w')
            Pickler(data).dump((self.subid,self.subscriptions))
            data.close()
        except (IOError,OSError):
            syslog(LOG_ERR,"Error writing statefile")

if __name__ == '__main__':
    try:
        (opts,args)=getopt(argv[1:],'d',['daemon='])
    except GetoptError,msg:
        print "%s\nUsage:\nemng.py [-d] [--daemon <pidfile>]"%(msg)
        exit(1)
    daemon=filter(lambda x:(x[0] == '--daemon'),opts)
    ldebug=len(filter(lambda x:(x[0] == '-d'),opts))
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    s=emng(debug=ldebug,statefile="/tmp/emng.pic")
    s.ServeForever()
